#!/usr/bin/env python3

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>

import logging
import psutil
import tempfile
import os
import sys
from .os import get_run_dir
import subprocess
from threading import Thread
from typing import List
import re

class PidDir(object):
    def __init__(
        self,
        tag    : str  = os.path.basename(sys.argv[0] if sys.argv[0] else __file__),
        delete : bool = False
    ):
        self._tag = tag
        self._delete = delete
        
        self.path = os.path.join(get_run_dir(), tag)
        
        if not os.path.exists(self.path):
            os.mkdir(self.path)
    
    def __del__(self):
        if self._delete:
            try:
                os.remove(self.path)
            except:
                logging.exception("Could not remove PidDir " + self._dir)

class PidFile(object):
    def __init__(
        self,
        dir : str = PidDir(delete=False).path,
        pid : int = os.getpid()
    ):
        self._dir = dir
        self._pid = pid
        
        fd, self.path = tempfile.mkstemp(suffix='.pid', prefix=str(pid), dir=dir, text=False)
        os.write(fd, '{}'.format(pid).encode())
        os.close(fd) # on windows if a file is still open it cannot be read by another process so close it and delete it later
        self._deleted = False
        logging.debug("Wrote pid file " + self.path)
    
    def remove(self):
        if not self._deleted:
            logging.debug("Removing pid file " + self.path)
            os.unlink(self.path)
            self._deleted = True
        
    def __del__(self):
        self.remove()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.remove()

class PidFileMonitor(object):
    def __init__(
        self,
        dir : str = PidDir(delete=False).path,
    ):
        self._dir = dir
    
    def running(self):
        if not os.path.isdir(self._dir):
            logging.debug("Pid dir " + self._dir + " does not exist.")
            return None
        else:
            def read_pid(pid_file : str):
                logging.debug("Reading pid file " + pid_file)
                
                with open(pid_file, "r") as f:
                    pid = f.read()
                try:
                    return int(pid)
                except ValueError as e:
                    logging.exception("Unexpected pid file " + pid_file)
                    return None
            
            pid_files = [os.path.join(self._dir, file)
                         for file in os.listdir(self._dir) 
                         if os.path.isfile(os.path.join(self._dir, file)) and
                            os.path.splitext(file)[1] == ".pid"]
            
            return [pid for pid in [read_pid(path) for path in pid_files] if pid is not None]
    
    def wait_all(self):
        for pid in self.running():
            try:
                proc = psutil.Process(pid)
            except psutil.NoSuchProcess:
                logging.info("No process with pid " + str(pid) + " found.")
            
            logging.info("Waiting for process with pid " + str(pid))
            while proc.is_running():
                try:
                    proc.wait(5)
                    logging.info("Still waiting...")
                except psutil.TimeoutExpired:
                    pass

class TeeThread(Thread):
    def __init__(self, in_file, *out_files, **kwargs):
        super(TeeThread, self).__init__(**kwargs)
        self._in_file = in_file
        self._out_files = out_files
        
        self.is_failed = False
        self.error = None
        self.traceback = None
    
    def run(self):
        self.is_failed = False
        self.error = None
        self.traceback = None
        try:
            for line in self._in_file:
                for f in self._out_files:
                    f.write(line)
        except:
            logging.debug("Writing to file or pipe failed", exc_info=True)
            self.is_failed = True
        finally:
            self.error = sys.exc_info()[1]
            self.traceback = sys.exc_info()[2]

def run_live(*popenargs, check=False, **kwargs):
    if ('stdout' in kwargs) or ('stderr' in kwargs) or ('bufsize' in kwargs) or ('universal_newlines' in kwargs):
        raise ValueError('stdout, stderr, bufsize and universal_newlines arguments may not be used')
    
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.PIPE
    kwargs['bufsize'] = 1
    kwargs['universal_newlines'] = True
    
    # References:
    #  https://github.com/python/cpython/blob/master/Lib/subprocess.py
    #  https://stackoverflow.com/a/4985080/6490710
    #  https://stackoverflow.com/a/17698359/6490710
    
    def tee(in_file, *out_files):
        t = TeeThread(in_file, *out_files, daemon=True)
        t.start()
        return t
    
    class StringWriter(object):
        def __init__(self):
            self.text = ''
            
        def write(self, text):
            self.text += text
    
    stdout = StringWriter()
    stderr = StringWriter()
    
    logging.info('$> {}'.format(*popenargs))
    
    with subprocess.Popen(*popenargs, **kwargs) as p:
        try:
            threads = []
            threads.append(tee(p.stdout, stdout, sys.stdout))
            threads.append(tee(p.stderr, stderr, sys.stderr))
            
            for t in threads: t.join()
            
            for t in threads:
                if t.is_failed:
                    raise subprocess.SubprocessError('logging thread(s) failed: {}\n{}'.format(t.error, t.traceback))
            
            if p.poll() is None:
                raise subprocess.SubprocessError('stdout/stderr logging threads stopped, but process is alive')
        except:
            p.kill()
            # No p.wait() here as __exit__ does that for us
            raise
        
        assert(p.poll() is not None) # None := process is alive
        retcode = p.poll()
        
        if check and retcode:
            raise subprocess.CalledProcessError(retcode, p.args, output=stdout.text, stderr=stderr.text)
        
        return subprocess.CompletedProcess(p.args, retcode, stdout.text, stderr.text)

def list_processes(names : List[str] = None):
    all = psutil.process_iter()
    if not names:
        return all
    
    regexes = [re.compile(name) for name in names]
    procs = []
    for proc in all:
        for regex in regexes:
            if regex.fullmatch(proc.name()):
                procs.append(proc)
    return procs



#!/usr/bin/env python3

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>

import os
import subprocess
import json
import logging
import sys

if sys.platform.startswith('linux'):
    def get_run_dir(uid : int = os.getuid()):
        run_dir = os.path.join("/run/user", str(uid))
        if not os.path.exists(run_dir) and uid == 0:
            run_dir = "/run"
        
        return run_dir

class FilesystemInfo(object):
    def __init__(self, uuid=None, name=None, fstype=None, mountpoint=None, label=None):
        self._uuid = uuid
        self._name = name
        self._fstype = fstype
        self._mountpoint = mountpoint
        self._label = label

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        if self._uuid != uuid:
            self._uuid = uuid
    
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if self._name != name:
            self._name = name

    @property
    def fstype(self):
        return self._fstype

    @fstype.setter
    def fstype(self, fstype):
        if self._fstype != fstype:
            self._fstype = fstype

    @property
    def mountpoint(self):
        return self._mountpoint

    @mountpoint.setter
    def mountpoint(self, mountpoint):
        if self._mountpoint != mountpoint:
            self._mountpoint = mountpoint
    
    @property
    def label(self):
        return self._label

    @mountpoint.setter
    def label(self, label):
        if self._label != label:
            self._label = label

if sys.platform.startswith('linux'):
    def list_filesystems():
        p = subprocess.run(["lsblk", "--json", "--list", "--fs"],
                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, 
                        check=True, universal_newlines=True)
        assert type(p) is subprocess.CompletedProcess
        
        class FilesystemInfoDecoder(json.JSONDecoder):
            def __init__(self, *args, **kargs):
                json.JSONDecoder.__init__(self, object_hook=self.dict_to_object, *args, **kargs)
            
            def dict_to_object(self, dct):
                if 'blockdevices' not in dct:
                    return dct
                
                fs_lst = dct.pop('blockdevices')
                try:
                    fs_nfo_lst = []
                    for fs in fs_lst:
                        uuid       = fs["uuid"]       if "uuid"       in fs else None
                        name       = fs["name"]       if "name"       in fs else None
                        fstype     = fs["fstype"]     if "fstype"     in fs else None
                        mountpoint = fs["mountpoint"] if "mountpoint" in fs else None
                        label      = fs["label"]      if "label"      in fs else None
                        
                        fs_nfo = FilesystemInfo()
                        fs_nfo.uuid       = uuid
                        fs_nfo.name       = name
                        fs_nfo.fstype     = fstype
                        fs_nfo.mountpoint = mountpoint
                        fs_nfo.label      = label
                        
                        fs_nfo_lst.append(fs_nfo)
                    return fs_nfo_lst
                except:
                    logging.exception('unexpected json')
                    dct['blockdevices'] = fs_lst
                    return dct
        
        fs_nfo_lst = json.loads(p.stdout, cls=FilesystemInfoDecoder)
        assert(type(fs_nfo_lst) is list and len(fs_nfo_lst) > 1 and type(fs_nfo_lst[0]) is FilesystemInfo)
        return fs_nfo_lst



#!/usr/bin/env python3

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>

import logging
import datetime
import os
import sys
import subprocess
from xml.etree import ElementTree
from .os import list_filesystems
from typing import List
import shutil
from .process import run_live
import json

if sys.platform.startswith('linux'):
    class SnapshotInfo(object):
        def __init__(
            self,
            path : str = None,
            type_ : str = None,
            num : int = None,
            date : datetime.datetime = None,
            description : str = None,
            cleanup : str = None
        ):
            self._path = path
            self._type_ = type_
            self._num = num
            self._date = date
            self._description = description
            self._cleanup = cleanup

        def __str__(self):
            return '{{path:{}; type:{}; num:{}; date:{}; description:{}; cleanup:{}}}'.format(
                self._path, self._type_, self._num, self._date, self._description, self._cleanup)
        
        @property
        def path(self):
            return self._path

        @path.setter
        def path(self, path):
            if self._path != path:
                self._path = path
        
        @property
        def type_(self):
            return self._type_

        @type_.setter
        def type_(self, type_):
            if self._type_ != type_:
                self._type_ = type_
                
        @property
        def num(self):
            return self._num

        @num.setter
        def num(self, num):
            if self._num != num:
                self._num = num
                
        @property
        def date(self):
            return self._date

        @date.setter
        def date(self, date):
            if self._date != date:
                self._date = date
                
        @property
        def description(self):
            return self._description

        @description.setter
        def description(self, description):
            if self._description != description:
                self._description = description
        
        @property
        def cleanup(self):
            return self._cleanup

        @cleanup.setter
        def cleanup(self, cleanup):
            if self._cleanup != cleanup:
                self._cleanup = cleanup
    
    class SnapshotBackupInfo(object):
        def __init__(
            self,
            mountpoint      : str,
            backup_dir      : str,
            tag             : str       = None,
            includes        : List[str] = None,
            excludes        : List[str] = None
        ):
            self._mountpoint = mountpoint
            self._backup_dir = backup_dir
            self._tag = tag
            self._includes = includes
            self._excludes = excludes

        def __str__(self):
            return '{{mountpoint:{}; type:{}; tag:{}; includes:{}; excludes:{}}}'.format(
                self._mountpoint, self._backup_dir, self._tag, self._includes, self._excludes)
        
        @property
        def mountpoint(self):
            return self._mountpoint

        @mountpoint.setter
        def mountpoint(self, mountpoint):
            if self._mountpoint != mountpoint:
                self._mountpoint = mountpoint
        
        @property
        def backup_dir(self):
            return self._backup_dir

        @backup_dir.setter
        def backup_dir(self, backup_dir):
            if self._backup_dir != backup_dir:
                self._backup_dir = backup_dir
                
        @property
        def tag(self):
            return self._tag

        @tag.setter
        def tag(self, tag):
            if self._tag != tag:
                self._tag = tag
                
        @property
        def includes(self):
            return self._includes

        @includes.setter
        def includes(self, includes):
            if self._includes != includes:
                self._includes = includes
                
        @property
        def excludes(self):
            return self._excludes

        @excludes.setter
        def excludes(self, excludes):
            if self._excludes != excludes:
                self._excludes = excludes
    
    class SnapshotBackupInfoEncoder(json.JSONEncoder):
        def default(self, obj):
            if not isinstance(obj, SnapshotBackupInfo):
                # Let the base class default method raise the TypeError
                return json.JSONEncoder.default(self, obj)
            
            return {
                '__type__'   : 'pywana.snapper.SnapshotBackupInfo',
                'mountpoint' : obj.mountpoint,
                'backup_dir' : obj.backup_dir,
                'tag'        : obj.tag,
                'includes'   : json.dumps(obj.includes),
                'excludes'   : json.dumps(obj.excludes)
            }
    
    def read_snapshot_info(snapshot_path : str):
        info_xml_path = os.path.join(snapshot_path, 'info.xml')
        info_xml = ElementTree.parse(info_xml_path)
        root = info_xml.getroot()
        
        path        = snapshot_path
        type_       = root.find('type').text
        num         = root.find('num').text
        date        = root.find('date').text
        description = root.find('description').text
        cleanup     = root.find('cleanup').text
        
        date = datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S") if date is not None else None
        return SnapshotInfo(path, type_, num, date, description, cleanup)
    
    
    def list_snapshots(mountpoint : str):
        snp_lst = []
        snp_parent_dir = os.path.join(mountpoint, '.snapshots')
        for snp_dir in os.listdir(snp_parent_dir):
            snp_path = os.path.join(snp_parent_dir, snp_dir)
            snp_lst.append(read_snapshot_info(snp_path))
        return snp_lst

    class InvalidMountpointError(Exception):
        pass
    
    class NotALinkError(Exception):
        pass
    
    def backup_snapshot(
        from_mountpoint : str,
        to_dir          : str,
        tag             : str       = None,
        includes        : List[str] = None,
        excludes        : List[str] = None,
        dry_run         : bool      = False,
        overwrite       : bool      = False
    ):
        fs = next((fs for fs in list_filesystems() if fs.mountpoint == from_mountpoint), None)
        
        if fs is None:
            raise InvalidMountpointError('\'{}\' is not a mountpoint'.format(from_mountpoint))
        
        if fs.fstype != 'btrfs':
            logging.warning('{} is not a btrfs filesystem'.format(from_mountpoint))
        
        if not os.path.isdir(to_dir):
            raise NotADirectoryError('{} is not a directory'.format(to_dir))
        
        if not os.access(to_dir, os.W_OK):
            raise PermissionError('{} is not writeable'.format(to_dir))
        
        if tag is None:
            if fs.uuid is None:
                raise ValueError('No tag given and filesystem {} has no valid uuid'.format(from_mountpoint))
            
            logging.debug('tag is {}'.format(fs.uuid))
            tag = fs.uuid
        
        snp_lst = list_snapshots(from_mountpoint)
        if not snp_lst:
            raise FileNotFoundError('No snapshots found for {}'.format(from_mountpoint))
        
        logging.debug('Found snapshots {}'.format([str(x) for x in snp_lst]))
        
        # take latest snapshot
        lst_by_date = sorted(snp_lst, key=lambda snp: snp.date)
        snp = lst_by_date[-1]
        
        bak_name = 'snapshot_of_uuid_{}_num_{}_date_{}'.format(fs.uuid, snp.num, snp.date.strftime('%Y%m%d%H%M%S'))
        bak_path = os.path.join(to_dir, bak_name)
        
        last_bak_name = 'snapshot_of_uuid_{}_LAST'.format(fs.uuid)
        last_bak_path = os.path.join(to_dir, last_bak_name)
        
        if os.path.exists(bak_path):
            if not overwrite:
                logging.warning('Backup dir {} already exists'.format(bak_path))
                return False
            logging.warning('Overwriting existing backup dir {}'.format(bak_path))
        
        if os.path.lexists(last_bak_path) and not os.path.islink(last_bak_path):
            raise NotALinkError('{} is not a link'.format(last_bak_path))
        
        if os.path.lexists(last_bak_path) and not os.path.exists(last_bak_path):
            logging.warning('Removing broken symlink {}'.format(last_bak_path))
            if not dry_run:
                os.unlink(last_bak_path)
        
        logging.info('Doing backup of snapshot {} to {}'.format(snp, bak_path))
        
        # make backup dir
        assert(overwrite if os.path.exists(bak_path) else True)
        if not os.path.exists(bak_path):
            if not dry_run:
                os.mkdir(bak_path)
        
        # write backup info first
        bak_nfo_path = os.path.join(bak_path, 'backup_info.json')
        bak_nfo = SnapshotBackupInfo(from_mountpoint, to_dir, tag, includes, excludes)
        bak_nfo_json = json.dumps(bak_nfo, cls=SnapshotBackupInfoEncoder, 
                                  sort_keys=True, indent=4, separators=(',', ': '))
        logging.debug('Writing backup info to {}: {}'.format(bak_nfo_path, bak_nfo_json))
        if not dry_run:
            with open(bak_nfo_path, 'w' if overwrite else 'x') as f:
                f.write(bak_nfo_json)
        
        # run rsync commands next
        default_cmd = ['rsync', '--stats', '--human-readable', '-vaHAX', '--one-file-system']
        
        if dry_run:
            default_cmd.append('--dry-run')
        
        def log_run(args):
            try:
                p = run_live(args,check=True)
            except subprocess.CalledProcessError as e:
                logging.exception('Command failed with exit code {}'.format(e.returncode))
                raise
            assert(type(p) is subprocess.CompletedProcess)
            logging.info('Command exited with code {}'.format(p.returncode))
            logging.debug('stdout:\n{}'.format(p.stdout))
            logging.debug('stderr:\n{}'.format(p.stderr))
        
        # rsync snapshot metadata
        meta_cmd = default_cmd.copy()
        meta_cmd.extend(['{}/'.format(snp.path), '{}/'.format(bak_path)])
        meta_cmd.extend(['--exclude', '/snapshot'])
        log_run(meta_cmd)
        
        # rsync snapshot data
        data_cmd = default_cmd.copy()
        data_cmd.extend(['{}/snapshot/'.format(snp.path), '{}/snapshot/'.format(bak_path)])
        
        if os.path.exists(last_bak_path):
            data_cmd.append('--link-dest={}/'.format(last_bak_path))
        
        if includes:
            for include in includes:
                data_cmd.extend(['--include', include])
        
        if excludes:
            for exclude in excludes:
                data_cmd.extend(['--exclude', exclude])
        
        log_run(data_cmd)
        
        # renew symlink to last backup
        if not dry_run:
            if os.path.lexists(last_bak_path):
                os.unlink(last_bak_path)
            os.symlink(bak_path, last_bak_path, target_is_directory=True)
        logging.debug('Created new symlink from {} to {}'.format(bak_path, last_bak_path))
        
        return True



#!/usr/bin/env python3

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>

from typing import List, Callable, Any
import signal
import logging
from threading import Event

class SignalHandler(object):
    def __init__(
        self, 
        signalnums : List[int],
        handler : Callable[[Any, Any], None]
    ):
        self._signalnums = signalnums
        self._handler = handler
    
    def register(self):
        def signal_handler(signum, frame):
            # Printing, e.g. by logging to stderr or stdout is not safe in signal handlers!
            # Ref.: https://bugs.python.org/issue24283
            self._handler(signum, frame)
    
        # set signal handler and store old handlers for later restore
        self._handlers = []
        for signalnum in self._signalnums:
            handler = signal.signal(signalnum, signal_handler)
            self._handlers.append([signalnum, handler])
        
        logging.debug('Registered signal handlers for ' + str(self._signalnums))
        return self
    
    def unregister(self):
        # reset handlers
        for signalnum,handler in self._handlers:
            signal.signal(signalnum, handler)
        logging.debug('Unregistered signal handlers for ' + str(self._signalnums))
    
    def __enter__(self):
        self.register()
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.unregister()

def wait_for_signals(signalnums : List[int], Timeout=None):
    fired = Event()
    def signal_handler(signum, frame):
        # Printing, e.g. by logging to stderr or stdout is not safe in signal handlers!
        # Ref.: https://bugs.python.org/issue24283
        fired.set()
    
    with SignalHandler(signalnums, signal_handler):
        logging.debug("Waiting for signals " + str(signalnums) + "...")
        return fired.wait(Timeout)



#!/usr/bin/env python3

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>


import logging
import dbus
import sys
import os
from gi.repository import GLib # from debian package python3-gi
from typing import List, Callable, Any
from threading import Thread

class LoginDProperties(object):
    def __init__(
        self,
        inhibit_delay_max_usec=None,
        preparing_for_shutdown=None,
        kill_user_processes=None
    ):
        self._inhibit_delay_max_usec = inhibit_delay_max_usec
        self._preparing_for_shutdown = preparing_for_shutdown
        self._kill_user_processes = kill_user_processes

    @property
    def inhibit_delay_max_usec(self):
        return self._inhibit_delay_max_usec

    @inhibit_delay_max_usec.setter
    def inhibit_delay_max_usec(self, inhibit_delay_max_usec):
        if self._inhibit_delay_max_usec != inhibit_delay_max_usec:
            self._inhibit_delay_max_usec = inhibit_delay_max_usec
    
    @property
    def preparing_for_shutdown(self):
        return self._preparing_for_shutdown

    @preparing_for_shutdown.setter
    def preparing_for_shutdown(self, preparing_for_shutdown):
        if self._preparing_for_shutdown != preparing_for_shutdown:
            self._preparing_for_shutdown = preparing_for_shutdown
            
    @property
    def kill_user_processes(self):
        return self._kill_user_processes

    @kill_user_processes.setter
    def kill_user_processes(self, kill_user_processes):
        if self._kill_user_processes != kill_user_processes:
            self._kill_user_processes = kill_user_processes

class LoginDConnection(object):
    def __init__(self):
        assert dbus.get_default_main_loop() is not None
        self._bus = dbus.SystemBus()
        self._proxy = self._bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
        self._properties = dbus.Interface(self._proxy, dbus_interface='org.freedesktop.DBus.Properties')
    
    def get_property(self, property : str):
        try:
            return self._properties.Get("org.freedesktop.login1.Manager", property)
        except dbus.exceptions.DBusException:
            logging.exception("Failed to retrieve logind property " + property + " via dbus")
            raise
    
    def get_properties(self):
        return LoginDProperties(
            inhibit_delay_max_usec=self.get_property("InhibitDelayMaxUSec"),
            preparing_for_shutdown=self.get_property("PreparingForShutdown"),
            kill_user_processes=self.get_property("KillUserProcesses")
        )
    
    def connect_to_signal(self, signal_name, handler_function, dbus_interface=None):
        self._proxy.connect_to_signal(signal_name, handler_function, dbus_interface)

class _GlibMainLoopThread(Thread):
    def __init__(self):
        super(_GlibMainLoopThread, self).__init__()
        assert dbus.get_default_main_loop() is not None
        self.is_failed = False
        self.error = None
        self.traceback = None
        self._loop = GLib.MainLoop()
        
    def __before__(self):
        pass
    
    def __after__(self):
        pass
    
    def run(self):
        self.is_failed = False
        self.error = None
        self.traceback = None
        try:
            self.__before__()
            self._loop.run()
            self.__after__()
        except Exception as e:
            logging.exception("Exception raised.")
            self.is_failed = True
        except KeyboardInterrupt as e:
            logging.exception("KeyboardInterrupt raised")
            self.is_failed = True
        except SystemExit as e:
            logging.exception("SystemExit raised")
            self.is_failed = True
        except:
            logging.exception("Error raised.")
            self.is_failed = True
        finally:
            self.error = sys.exc_info()[1]
            self.traceback = sys.exc_info()[2]
    
    def stop(self):
        if self._loop is not None:
            self._loop.quit()



class InhibitLock(object):
    def __init__(
        self,
        what : str ="shutdown:sleep",
        who  : str = os.path.basename(sys.argv[0] if sys.argv[0] else __file__),
        why  : str = "",
        mode : str = "delay"
    ):
        self._what = what
        self._who = who
        self._why = why
        self._mode = mode
        assert dbus.get_default_main_loop() is not None
        
        if mode == "delay":
            inhibit_delay_max_usec = LoginDConnection().get_properties().inhibit_delay_max_usec
            if inhibit_delay_max_usec < 30 * 1000 * 1000:
                logging.warning("Maximum shutdown delay is very short, increase InhibitDelayMaxUSec= " +
                                "in logind.conf! Help: man logind.conf")
    
    def lock(self):
        bus = dbus.SystemBus()
        self._inhibit_lock = bus.call_blocking(
            'org.freedesktop.login1', # bus_name
            '/org/freedesktop/login1', # object_path
            'org.freedesktop.login1.Manager', # dbus_interface
            'Inhibit', # method
            'ssss', # signature
            (
                self._what, # what
                self._who, # who
                self._why, # why
                self._mode #mode
            ), # args
            timeout=2.0 # timeout=-1.0
            # byte_arrays=False
            # **kwargs
        )
        logging.debug('Aquired lock for ' + self._what)
    
    def unlock(self):
        del self._inhibit_lock
        logging.debug('Released lock for ' + self._what)
        
    def __enter__(self):
        self.lock()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.unlock()

class PrepareForShutdownMonitor(_GlibMainLoopThread):
    def __init__(
        self, 
        handler : Callable[[], None]
    ):
        super(PrepareForShutdownMonitor, self).__init__()
        self._handler = handler
        assert dbus.get_default_main_loop() is not None
        self._logind_proxy = LoginDConnection()
    
    def __before__(self):
        logging.debug("__before__")
        
        def prepare_for_shutdown_handler(active):
            logging.debug("PrepareForShutdown(" + str(active) + ") received.")
            assert active
            self._handler()
        
        self._logind_proxy.connect_to_signal("PrepareForShutdown", prepare_for_shutdown_handler)
    
    def __after__(self):
        logging.debug("__after__")
    
    def wait(self, Timeout=None):
        self.join(Timeout)
        return self.is_alive() == False



#!/usr/bin/env python3

# vim:set tabstop=8 shiftwidth=4 expandtab:
# kate: syntax python; space-indent on; indent-width 4;

# Copyright (c) 2019 Jakob Meng, <jakobmeng@web.de>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import setuptools
setuptools.setup()
